Imports System.IO
''' <summary>
''' Klasse mit Methoden zum Anlegen und Anzeigen von Thumbnails fr Dateien
''' </summary>
''' <remarks>Im Hintergrund luft ein Thread, der dauernd versucht, neue 
''' Thumbnails zu erstellen. Da dieser Thread fr eine Anwendung nur einmal
''' existieren soll, ist er Shared, die Klasse Pfad liefert auch nur Shared Methoden 
''' um z.B. neue Pfade zu erstellen.
''' 
''' Hier verwende ich folgendes Prinzip zum Erstellen von Thumbnails:
''' Die erzeugten Thumbnails sind maximal 400x400 Pixel gro.
''' Sie sind in einem temporren Directory mit gleich bleibendem Pfad 
''' gespeichert, das unverndert bleibt, solange die temporren Dateien 
''' nicht gelscht werden. Also bleiben Thumbnails auch nach Schlieen und 
''' Neu-ffnen der Anwendung erhalten. 
''' Um zu erkennen, da ein Thumbnail noch aktuell ist, speichere ich das 
''' Thumbnail-Image als PNG und nebenan eine XML-Datei mit Information zu
''' LastWriteTime und Byte-Lnge der Original-Image-Datei.
''' Der Pfad zum ThumbnailImage entspricht dem temporren Ordner-Namen
''' + dem realen Pfad der Original-Image-Datei + ".thmb" (zu finden in 
''' der Methode GetThumbPath). Damit sollte Eindeutigkeit fr Original-Bilder
''' und Thumbnails gesichert werden knnen.</remarks>
Public MustInherit Class ThumbPath

#Region " Private Felder "
    ' Pfad zur angezeigten Datei
    Private m_Path As String
#End Region

#Region " Properties / Events "
    ''' <summary>
    ''' Gibt an oder legt fest, ob zu diesem Pfad eine geeignete Thumbnail-Datei vorhanden ist
    ''' </summary>
    MustOverride Property HasImage() As Boolean
    ''' <summary>
    ''' Gibt an, ob es sich bei diesem Pfad um einen BildPfad handelt
    ''' </summary>
    MustOverride ReadOnly Property IsImagePath() As Boolean
    ''' <summary>
    ''' Event wird geliefert, wenn sich fr eine Bilddatei die Verfgbarkeit der Thumbnail-Datei
    ''' gendert hat
    ''' </summary>
    Public Event Image_Changed()
    ''' <summary>
    ''' erlaubt, das Changed-Event von erbenden Klassen aus zu liefern
    ''' </summary>
    Protected Sub RaiseImageChanged()
        RaiseEvent Image_Changed()
    End Sub
#End Region

#Region " Konstruktor "

    Private Sub New(ByVal Path As String)
        ' Privater Konstruktor, um zu verhindern, dass ein Objekt vom Typ Pfad 
        ' ohne die Shared-Methode Create angelegt werden knnte.
        m_Path = Path
    End Sub

#End Region

#Region " Methoden "
    Public Overrides Function ToString() As String
        Return m_Path
    End Function
#End Region


#Region " Shared "

#Region " Private  Felder "

    ' Liste mit Bild-Pfaden, fr die neue Thumbnails angelegt werden mssen
    Private Shared m_BilderToCreate As New List(Of BildPfad)
    ' Thread, der dauernd dafr sorgt da neue Thumbnails erzeugt werden
    Private Shared m_ThumbCreationThread As Threading.Thread
    ' Flag zum Mitteilen an den ThumbCreationThread, dass jetzt 
    ' neue Thumbnails erzeugt werden mssen
    Private Shared m_MustResetThumbCreation As Boolean = False
    ' Root-Pfad im Temp-Ordner, unter dem alle Thumbnails in einer Ordnerstruktur
    ' gespeichert werden
    Private Shared m_Root As String

#End Region

#Region " Properties "

#End Region

#Region " Konstruktor "
    Shared Sub New()
        m_Root = IO.Path.GetTempPath & "ThumbnailCreator\"
    End Sub
#End Region

#Region " Methoden "

    ''' <summary>
    ''' liefert den Thumbnail-Pfad fr eine Bilddatei 
    ''' </summary>
    ''' <param name="FileName">Dateiname der Bilddatei</param>
    ''' <returns>Pfadangabe im temporren Ordner</returns>
    Public Shared Function GetThumbPath(ByVal FileName As String)
        Return m_Root & FileName.Replace(":", "") & ".thmb"
    End Function

    ''' <summary>
    ''' Liefert ein Array mit Pfad-Objekten fr als String-Array angegebene 
    ''' Dateipfade
    ''' </summary>
    ''' <param name="Paths">Dateipfade</param>
    ''' <returns>Array mit Pfad-Objekten</returns>
    Shared Function Create(ByVal Paths() As String) As ThumbPath()
        ' Liste fr Resultate
        Dim Result As New List(Of ThumbPath)
        ' aus dieser Auflistung erhlt der ThumbCreationThread neue Pfade
        m_BilderToCreate.Clear()

        ' fr alle angegebenen Datei-Pfade
        For Each txt As String In Paths
            Dim Pfd As ThumbPath = Nothing

            If IsImage(txt) Then
                ' wenn die Datei eine Bilddatei ist, schnell untersuchen,
                ' ob eine korrekte Thumbnail-Datei existiert.
                ' wenn die Thumbnail-Datei nicht mehr der aktuellen Bilddatei entspricht
                ' (z.B. die Bilddatei zwischenzeitlich verndert wurde),
                ' wird bei ThumbExists diese Datei gelscht
                Dim ThumbExists As Boolean = CheckThumbExists(txt)
                Pfd = New BildPfad(txt, ThumbExists)
                If Not ThumbExists Then
                    ' wenn noch kein Thumbnail existiert, diesen Pfad zur Liste 
                    ' neu zu erstellender Thumbnails hinzufgen
                    m_BilderToCreate.Add(Pfd)
                End If
            Else
                ' wenn der Pfad zu keiner Bilddatei gehrt, braucht auch nicht
                ' ein Thumbnail erstellt zu werden
                Pfd = New NormalPfad(txt)
            End If
            ' Neuen Pfad zur Ergebnis-Liste hinzu fgen
            Result.Add(Pfd)
        Next

        ' hier Flag fr ThumbCreationThread auf True setzen, damit dieser sich jetzt
        ' um die neuen Bildpfade kmmert
        m_MustResetThumbCreation = True
        ' Ergebnis zurck geben
        Return Result.ToArray
    End Function

    ''' <summary>
    ''' untersucht fr einen Datei-Pfad, ob es sich um eine Bilddatei handelt
    ''' </summary>
    ''' <param name="Path">Datei-Pfad</param>
    Private Shared Function IsImage(ByVal Path As String) As Boolean
        Dim Extension As String = IO.Path.GetExtension(Path).ToLower
        If Extension = ".png" Then Return True
        If Extension = ".jpg" Then Return True
        If Extension = ".gif" Then Return True
        Return False
    End Function

    ''' <summary>
    ''' Startet den ThumbCreationThread
    ''' </summary>
    Public Shared Sub StartThreading()
        m_ThumbCreationThread = New Threading.Thread(AddressOf CreateAllThumbs)
        m_ThumbCreationThread.Start()
    End Sub

    ''' <summary>
    ''' Beendet den ThumbCreationThread
    ''' </summary>
    Public Shared Sub AbortThreading()
        If m_ThumbCreationThread IsNot Nothing Then
            m_ThumbCreationThread.Abort()
            m_ThumbCreationThread.Join()
            m_ThumbCreationThread = Nothing
        End If
    End Sub

    ''' <summary>
    ''' Methode, die dauernd neue Thumbnails erzeugt
    ''' </summary>
    Private Shared Sub CreateAllThumbs()

        ' Liste mit Pfaden fr zu erzeugende Thumbnails
        Dim m_CreationList As New List(Of BildPfad)

        ' unendliche Schleife
        While True
            ' nur wenn MustResetThumbCreation-Flag True ist, 
            ' wird die Liste zum Erzeugen neuer Thumbnails aktualisiert.
            If m_MustResetThumbCreation Then
                ' Flag gleich wieder auf False setzen
                m_MustResetThumbCreation = False
                ' Liste neu fllen
                m_CreationList.Clear()
                m_CreationList.AddRange(m_BilderToCreate.ToArray)
            End If

            ' solange in der CreationList Elemente enthalten sind, 
            ' neuen Thumbnails erzeugen
            While m_CreationList.Count > 0
                Dim ActiveBild As BildPfad = m_CreationList(0)
                Try
                    ' Thumbnail erstellen
                    CreateThumb(ActiveBild.ToString)
                    m_CreationList.RemoveAt(0)
                    ActiveBild.HasImage = True
                Catch ex As Exception
                    ' Bei einem Fehler, z.B. IOException
                    ' wird der Pfad fr einen nochmaligen Versuch 
                    ' an das hintere Ende der Liste bewegt
                    m_CreationList.RemoveAt(0)
                    m_CreationList.Add(ActiveBild)
                    ' nur zum Testen
                    Debug.WriteLine("Fehler! " & ex.Message)
                End Try
                ' wenn MustResetThumbCreation-Flag True ist, diese
                ' While-Schleife verlassen
                If m_MustResetThumbCreation Then
                    Exit While
                End If
            End While
        End While

    End Sub

    ''' <summary>
    ''' Erzeugt eine neue Thumbnail-Datei fr einen angegebenen Bild-Pfad
    ''' </summary>
    ''' <param name="OriginalPath">Pfad zur Original-Datei</param>
    Private Shared Sub CreateThumb(ByVal OriginalPath As String)
        Dim CP As String = OriginalPath
        ' Pfad, unter dem die Thumbnail-Datei gespeichert werden soll, ermitteln
        Dim ThumbPath As String = GetThumbPath(CP)
        ' FileInfo, fr die Original-Datei,
        ' enthlt unter anderem LastWriteTime und Lnge der Datei in Bytes
        Dim FI As New IO.FileInfo(CP)
        Dim FileDate As DateTime = FI.LastWriteTime
        Dim FileLength As Long = FI.Length

        ' ein Klon der Original-Image Datei erzeugen und gleich wieder frei geben
        Dim Img As Image
        Using TmpImg As Image = Image.FromFile(OriginalPath)
            Img = TmpImg.Clone
        End Using

        ' Kantenabmessungen des neuen Thumbnail-Bildes (max. 400x400 Pixel)
        Dim ThumbSize As Size
        If Img.Width > 400 OrElse Img.Height > 400 Then
            ThumbSize = GetThumbSize(Img.Size)
        Else
            ThumbSize = Img.Size
        End If

        ' neue Bitmap mit ermittelten Kantenabmessungen erstellen
        Using Thumb As New Bitmap(ThumbSize.Width, ThumbSize.Height, Imaging.PixelFormat.Format32bppArgb)
            ' auf dem Graphics-Objekt dieser Bitmap das Originalbild zeichnen
            Using g As Graphics = Graphics.FromImage(Thumb)
                g.DrawImage(Img, New Rectangle(0, 0, ThumbSize.Width, ThumbSize.Height))
            End Using

            ' Falls notwendig, Directory-Struktur fr die neue Thumbnail-Datei erzeugen
            Dim Dirname As String = Path.GetDirectoryName(ThumbPath)
            If Not IO.Directory.Exists(Dirname) Then
                IO.Directory.CreateDirectory(Dirname)
            End If

            ' Thumb-Datei als PNG speichern
            Thumb.Save(ThumbPath, Drawing.Imaging.ImageFormat.Png)

            ' Wesentliche Eigenschaften des Original-Bildes in einer XML-Datei festhalten
            ' um spter feststellen zu knnen, ob die Thumb-Datei noch aktuell ist
            Dim xWriter As New Xml.XmlTextWriter(ThumbPath & ".xml", System.Text.Encoding.UTF8)

            xWriter.WriteStartDocument()
            xWriter.WriteStartElement("Elemente")
            xWriter.WriteAttributeString("LastWriteTime", FileDate.Ticks.ToString)
            xWriter.WriteAttributeString("FileLength", FileLength.ToString)
            xWriter.WriteEndDocument()
            xWriter.Close()

        End Using
        Img.Dispose()
    End Sub

    ''' <summary>
    ''' Liefert passende Kantenabmessungen fr eine Thumbnail-Datei
    ''' </summary>
    ''' <param name="OriginalImageSize">Abmessungen der Original-Bilddatei</param>
    ''' <remarks>(die lngere Seite soll 400 Pixel lang sein)</remarks>
    Private Shared Function GetThumbSize(ByVal OriginalImageSize As Size) As Size
        Dim pw As Single = OriginalImageSize.Width
        Dim ph As Single = OriginalImageSize.Height
        Dim dw As Single = 400.0F
        Dim dh As Single = 400.0F

        Dim scaleFactor As Single
        If pw / ph > dw / dh Then
            scaleFactor = dw / pw
        Else
            scaleFactor = dh / ph
        End If

        Dim NewWidth As Integer = Math.Min(scaleFactor * pw, 400)
        Dim NewHeight As Integer = Math.Min(scaleFactor * ph, 400)
        Return New Size(NewWidth, NewHeight)
    End Function

    ''' <summary>
    ''' liefert eine Information, ob eine zu einer Original-Datei passende Thumb-Datei
    ''' vorhanden ist und lscht falls notwendig eine nicht mehr aktuelle Thumb-Datei
    ''' </summary>
    ''' <param name="FileName">Pfad zur Original-Bilddatei</param>
    Private Shared Function CheckThumbExists(ByVal FileName As String) As Boolean
        Dim ThumbFileName As String = GetThumbPath(FileName)

        If IO.File.Exists(ThumbFileName & ".xml") Then
            Dim Reader As New Xml.XmlTextReader(ThumbFileName & ".xml")
            Reader.MoveToContent()
            Reader.MoveToAttribute("LastWriteTime")
            Dim ThumbDate As New DateTime(CLng(Reader.Value))
            Reader.MoveToAttribute("FileLength")
            Dim ThumbLength As Long = CLng(Reader.Value)
            Reader.Close()

            Dim FI As New IO.FileInfo(FileName)
            ' Datum und Datei-Gre mit den entsprechenden Angaben der Original-Bilddatei
            ' vergleichen
            If ThumbDate = FI.LastWriteTime AndAlso ThumbLength = FI.Length Then
                Return True
            Else
                ' wenn die Thumb-Datei nicht zur Original-Bilddatei passt,
                ' wird die Thumb-Datei gelscht
                IO.File.Delete(ThumbFileName)
                IO.File.Delete(ThumbFileName & ".xml")
            End If
        Else
            If IO.File.Exists(ThumbFileName) Then
                IO.File.Delete(ThumbFileName)
            End If
        End If

        Return False
    End Function

#End Region

#End Region

#Region " Private Klassen "

#Region " BildPfad "
    ''' <summary>
    ''' Klasse mit Pfadangabe fr eine Bilddatei
    ''' </summary>
    Private Class BildPfad
        Inherits ThumbPath

#Region " Private Felder "
        ' Enthlt Information, ob eine korrekte Thumbnail-Datei vorhanden ist
        Private m_HasImage As Boolean = False
#End Region

#Region " Properties "

        ''' <summary>
        ''' Gibt an oder legt fest, ob zu diesem Pfad eine geeignete Thumbnail-Datei vorhanden ist
        ''' </summary>
        Public Overrides Property HasImage() As Boolean
            Get
                Return m_HasImage
            End Get
            Set(ByVal value As Boolean)
                If m_HasImage <> value Then
                    m_HasImage = value
                    MyBase.RaiseImageChanged()
                End If
            End Set
        End Property

        ''' <summary>
        ''' Gibt an, ob es sich bei diesem Pfad um einen BildPfad handelt
        ''' </summary>
        Public Overrides ReadOnly Property IsImagePath() As Boolean
            Get
                Return True
            End Get
        End Property
#End Region

#Region " Konstruktor "

        ''' <summary>
        ''' Konstruktor mit Pfad zur Original-Bilddatei und Angabe, ob 
        ''' eine korrekte Thumnail-Datei existiert
        ''' </summary>
        ''' <param name="Path">Pfad zur Original-Bilddatei</param>
        ''' <param name="ThumbExists">Angabe, ob korrekte Thumnail-Datei existiert</param>
        Public Sub New(ByVal Path As String, ByVal ThumbExists As Boolean)
            MyBase.New(Path)
            m_HasImage = ThumbExists
        End Sub

#End Region

#Region " Methoden "

#End Region

    End Class

#End Region

#Region " NormalPfad "

    Private Class NormalPfad
        Inherits ThumbPath

#Region " Properties "

        ''' <summary>
        ''' Gibt an, ob zu diesem Pfad eine geeignete Thumbnail-Datei vorhanden ist
        ''' </summary>
        Public Overrides Property HasImage() As Boolean
            Get
                Return True
            End Get
            Set(ByVal value As Boolean)

            End Set
        End Property

        ''' <summary>
        ''' Gibt an, ob es sich bei diesem Pfad um einen BildPfad handelt
        ''' </summary>
        Public Overrides ReadOnly Property IsImagePath() As Boolean
            Get
                Return False
            End Get
        End Property
#End Region

#Region " Konstruktor "

        Public Sub New(ByVal Path As String)
            MyBase.New(Path)
        End Sub

#End Region

#Region " Methoden "

#End Region

    End Class
#End Region

#End Region

End Class
